-- ********************************************************************************
-- *  Copyright (C) Thorsten Gaertner, Oststeinbek / Germany 2004                 *
-- ********************************************************************************
-- *  This source file may be used and distributed without restriction            *
-- *  provided that this copyright statement is not removed from the file and     *
-- *  that any derivative work contains the original copyright notice and the     *
-- *  associated disclaimer.                                                      *
-- ********************************************************************************
-- *  This source file is provided "as is" and WITHOUT any express or implied     *
-- *  warranties, that this source file is                                        *
-- *  1. free from any claims of infringement,                                    *
-- *  2. the merchantability or fitness for a particular purpose.                 *
-- ********************************************************************************
--
-- ********************************************************************************
-- *  File        : VCO_WITH_PLL.vhd                                              *
-- *  Version     : 1.0                                                           *
-- *  Target      : Plain VHDL (only for simulation, NOT for synthesis)           *
-- *  Description : Voltage Controlled Oscillator with Phase Locked Loop          *
-- ********************************************************************************

library IEEE;
use     IEEE.STD_LOGIC_1164.all;
use     IEEE.STD_LOGIC_ARITH.all;
use     IEEE.STD_LOGIC_UNSIGNED.all;

library WORK;
use     WORK.CONVERSION_PCK.all;

-- ********************************************************************************

entity VCO_WITH_PLL is
  generic (
    VCO_MIN_FREQ  : real :=   1000000.0;  -- VCO Minimum Frequency [Hz]
    VCO_MAX_FREQ  : real := 100000000.0;  -- VCO Maximum Frequency [Hz]
    VCO_BEG_FREQ  : real :=  10000000.0;  -- VCO Start Frequency at time = 0 [Hz]
    DAMPING       : real :=       0.001   -- Damping Factor from the Phase Comparator to the VCO-Control-Level
  );
  port (
    CLK_REF_IN    : in  std_logic;  -- Reference Clock Input
    CLK_FB_IN     : in  std_logic;  -- Reference Clock Input
    CLK_VCO_OUT   : out std_logic   -- Output Clock of the VCO
  );
end VCO_WITH_PLL;

-- ********************************************************************************

architecture VCO_WITH_PLL of VCO_WITH_PLL is

-- ********************************************************************************

signal  vco_actual_frequency : real := VCO_BEG_FREQ;                                                   -- Frequency of the Oscillation of the VCO
signal  vco_actual_period    : time := (1.0 / vco_actual_frequency) * 1 sec;                           -- Period Time for the Oscillation of the VCO
signal  vco_level            : real := (VCO_BEG_FREQ - VCO_MIN_FREQ) / (VCO_MAX_FREQ - VCO_MIN_FREQ);  -- "Control-Voltage" for the VCO. Range 0..1

signal  period_fb , time_fb  : real := 1.0;
signal  period_ref, time_ref : real := 1.0;
signal  pc                   : real := 0.0;

signal  clk_fb_in_calc       : std_logic;

-- ********************************************************************************

begin

-- ********************************************************************************
-- **       Detection of rising edges for calculation of the period times        **
-- ********************************************************************************

phase_fb_proc : process
variable time_fb_now : real;
begin
  wait until rising_edge (CLK_FB_IN);

  time_fb_now := CONV_TIME_TO_REAL (now);  -- time : now
  period_fb   <= time_fb_now - time_fb;
  time_fb     <= time_fb_now;

end process;

-- ********************************************************************************

phase_ref_proc : process
variable time_ref_now : real;
begin
  wait until rising_edge (CLK_REF_IN);

  time_ref_now := CONV_TIME_TO_REAL (now);  -- time : now
  period_ref   <= time_ref_now - time_ref;
  time_ref     <= time_ref_now;

end process;

-- ********************************************************************************

clk_fb_in_calc <= transport CLK_FB_IN after 1 ps;

-- ********************************************************************************
-- **                       Calculate the new "vco_level"                        **
-- ********************************************************************************

calc_level_proc : process
variable phase_comp, vco_level_new : real;
begin
  wait until rising_edge (clk_fb_in_calc);

  if    (period_fb  > (2.0 * period_ref)) then  phase_comp := +1.0;  -- correction : ref-freq too fast / fb-freq too slow
  elsif (period_ref > (2.0 * period_fb )) then  phase_comp := -1.0;  -- correction : ref-freq too slow / fb-freq too fast

  elsif (period_fb  >         period_ref) then  phase_comp :=   2.0 * (1.0 - (period_ref / period_fb ));
  elsif (period_ref >         period_fb ) then  phase_comp := - 2.0 * (1.0 - (period_fb  / period_ref));

  else                                          phase_comp := 0.0;  -- ref-freq = fb-freq;
  end if;

  pc <= phase_comp;

  vco_level_new := vco_level + ( phase_comp * DAMPING );

  if    (vco_level_new < 0.0)  then  vco_level <= 0.0;
  elsif (vco_level_new > 1.0)  then  vco_level <= 1.0;
  else                               vco_level <= vco_level_new;
  end if;

end process;

-- ********************************************************************************
-- **                  Calculate the new period time for the VCO                 **
-- ********************************************************************************

vco_time_proc : process
begin
  wait for vco_actual_period;

  vco_actual_frequency <= VCO_MIN_FREQ + ( vco_level * (VCO_MAX_FREQ - VCO_MIN_FREQ) );

  vco_actual_period    <= (1.0 / vco_actual_frequency) * 1 sec;

end process;

-- ********************************************************************************
-- **                    V C O : Voltage Controlled Oscillator                   **
-- ********************************************************************************

vco_proc : process
begin
  wait for (vco_actual_period / 2);  CLK_VCO_OUT <= '0';
  wait for (vco_actual_period / 2);  CLK_VCO_OUT <= '1';
end process;

-- ********************************************************************************

end VCO_WITH_PLL;

-- ********************************************************************************
-- List Of Changes
-- ===============
-- DATE        VERSION  NAME  CHANGE
-- 25.09.2004  1.0      ThG   Creation
-- ********************************************************************************
